import java.applet.*;
import java.awt.*;
import java.io.*;
import java.util.*;
import java.net.*;
import java.awt.BorderLayout;
import java.awt.event.*;
import graphoutput.*;
import data.*;

/**
 * Main class of the NDimViewer-visualization-applet to visualize
 * n-dimensional Dynamical Systems. It contains the main menu consisting
 * of the functions "data", "phase-plot", "info", and "help".
 *
 * @author Peter Steiger
 * @version 1.0
 */
public class NDimViewer extends Applet implements DataSink, Reporter, ActionListener
{
   /**
    * The instance of the current active output-frame
    */
   public static GraphOutput go;

   private Button dataButton;
   private Button phaseButton;
   private Button helpButton;
   private Button infoButton;
   private Label  lblStatus;

   /**
    * In this object the resulting values of the current calculation are stored
    */
   DataList	data;
   /**
    * Contains start- and endtime of the current system
    */
   Range time;
   /**
    * Handle to the DataDialog. It is non-null, when the dialog is active.
    */
   DataDialog datadlg;


   /**
    * Calls the various methods necessary to initialize the
    * program.
    */
    public void init()
	{
      Panel p = new Panel();

      Panel up = new Panel();
	  up.setLayout(new FlowLayout());
	  dataButton = new Button("Data");
	  dataButton.addActionListener(this);
	  phaseButton = new Button("Phase-Plot");
	  phaseButton.addActionListener(this);
	  helpButton = new Button("Help");
	  helpButton.addActionListener(this);
	  infoButton = new Button("Info");
	  infoButton.addActionListener(this);
	  up.add(dataButton);
	  up.add(phaseButton);
	  up.add(helpButton);
	  up.add(infoButton);

	  lblStatus = new Label("Ready");

	  p.setLayout(new BorderLayout());
      p.add("North", up);
	  p.add("South", lblStatus);

	  add(p);
	  setVisible(true);
    }

	/**
	 *  Handles the event when a button of the main menu is pressed.
	 */
	public void actionPerformed(ActionEvent e)
    {
	  if(e.getActionCommand().equals(dataButton.getLabel()))
	  {
	    datadlg = new DataDialog( this );
		datadlg.show();
	  }
      else if(e.getActionCommand().equals(phaseButton.getLabel()))
	  {
		PhasenDlg dlg = new PhasenDlg( data );
		dlg.show();
	  }
      else if(e.getActionCommand().equals(infoButton.getLabel()))
	  {
		InfoDialog dlg = new InfoDialog( this );
		dlg.show();
	  }
      else if(e.getActionCommand().equals(helpButton.getLabel()))
	  {
		HelpWindow dlg = new HelpWindow(this, "help/main.hlp",1, "Help Main-Menu", 500, 400);
        dlg.show();
	  }

	}

    /**
	 *  Delivers the calculated data to the callee
	 */
	public DataList getData()
	{
	  return data;
	}

	/**
	 *  Sets the current data-list. After calculation is finished, the
     *  data is propagated to this class via this function.
	 */
	public void setData( DataList d )
	{
	  data = d;
	}

	/**
	 *  Sets the text of the statuslines
	 */
	public void setStatus( String s )
	{
	  lblStatus.setText(s);
      if(go!=null) go.setStatusLine(s);
	}

	/**
	 *  Sets the time range of this calculation
	 */
	public void setTime(double startTime, double endTime)
	{
	  time = new Range(startTime, endTime);
	}

	/**
	 * Does a new iteration with the specified startvalue_v
	 * and writes the result into this.data
	 */
	public void iterate (Vector startvalue_v)
	{
      DataDialog dlg = new DataDialog(this);

      for (int i=0; i<startvalue_v.size(); i++)
      {
        dlg.elementOrigin.set(i, (Double)startvalue_v.elementAt(i) );
      }

      data = dlg.parseSystem();

	}


	/**
	 *  Returns the information about this applet in a string
	 */
	public String getAppletInfo()
	{
		return "Name: NDimViewer\r\n" +
		       "Function: 3-dimensional displaying of high-dimensional\r\ndynamical systems\r\n" +
		       "Author: Peter Steiger\r\n" +
		       "Created with JDK 1.2 and Java3D 1.1";
	}



} // end of class nDimViewer