package graphoutput;

import javax.media.j3d.*;
import com.sun.j3d.utils.geometry.*;
import com.sun.j3d.utils.behaviors.picking.*;
import javax.vecmath.*;
import java.util.*;
import java.awt.*;

/**
 * This class contains the implementation of the method
 * "Extruded Parallel Coordinates". It creates the geometry within
 * this BranchGroup into the 1x1x1-box.
 */
public class EPCGraph extends BranchGroup
{
    private int dimcnt, pointcnt;
	private Vector stripe_v;  
	private Vector stripeShape_v;
	private LineArray lines;
	private Shape3D lineShape = new Shape3D();

	private Style style;
	private GraphOutput parent;
    private BranchGroup stripe_bg;

    /**
     * Removes the current geometry (stripes & lines)
     */
    public void removeGeometry()
    { 
      stripe_bg.detach();
      stripeShape_v = new Vector();
      lineShape = new Shape3D();
    }

    /**
     * Draws new stripes using the specified data
     */
    public void updateGeometry(Vector points_v, Vector scaling_v)
    {
      createGeometries(dimcnt, pointcnt, points_v, scaling_v);

      for (int i=0; i<dimcnt; i++)
        ((Shape3D)stripeShape_v.elementAt(i)).setAppearance(style.getAppearance(i));

	  if (style.showlines==true && dimcnt>1)
	  {
	    Appearance app1 = new Appearance();
	    Color3f color1 = new Color3f(1.0f, 1.0f, 1.0f);
	    ColoringAttributes ca1 = new ColoringAttributes();
	    ca1.setColor(color1);
	    app1.setColoringAttributes(ca1);

	    PolygonAttributes pa1 = new PolygonAttributes();
	    pa1.setPolygonMode(pa1.POLYGON_LINE);
	    pa1.setCullFace(pa1.CULL_NONE);
	    app1.setPolygonAttributes(pa1);
	    lineShape.setAppearance(app1);
	  }


      stripe_bg = new BranchGroup();
      stripe_bg.setCapability(BranchGroup.ALLOW_DETACH);

      for (int i=0; i<dimcnt; i++)
	    stripe_bg.addChild((Shape3D)stripeShape_v.elementAt(i));
	  if (style.showlines==true && dimcnt>1) stripe_bg.addChild(lineShape);

      addChild(stripe_bg);

    }


    /**
     * Fills the stripeShape_v with the stripes as Shape3Ds
     */
	private void createGeometries(int dimcnt, int pointcnt, Vector points_v, Vector scaling_v)
	{
      TriangleStripArray stripe;

      // show the coordinate-box(to get sure, if its hidden)
      parent.showCoordinateBox();

      stripe_v = new Vector();
      stripeShape_v = new Vector();

	  Vector3f normal = new Vector3f(0.0f, 0.0f, 0.0f);
	  double stripwidth = 0.0;
	  double spacewidth = 0.0;

      int vertexcnt = 2*pointcnt;
	  int[] stripelength;
	  stripelength = new int[1];
      stripelength[0]=2*pointcnt;

      if(style.showlines==true && dimcnt>1)
	  {
	    lines = new LineArray(2*(dimcnt-1)*pointcnt, LineArray.COORDINATES);
	  }

	  if(dimcnt>1)
	  {
	    stripwidth = (1-style.space_frac) / (double) dimcnt;
	    spacewidth = style.space_frac / ((double) dimcnt - 1.0);
	  }
	  else
	  {
	    stripwidth = 1.0;
	    spacewidth = 0.0;
	  }
      for(int d=0; d<dimcnt; d++)
      {
		stripe = new TriangleStripArray(vertexcnt,
	                                    TriangleArray.COORDINATES | TriangleArray.NORMALS,
									    stripelength);

	    for(int k=0; k<pointcnt; k++)
	    {
  		  double x1 = d*(stripwidth+spacewidth);
		  double x2 = x1 + stripwidth;
		  double y = ((Range)scaling_v.elementAt(d)).normalize(
		               ((Double)((Vector)points_v.elementAt(k)).elementAt(d)).doubleValue());
		  double z = (double) k / (double) (pointcnt-1);
		  double[] p = { x1, y, z, x2, y, z };

		  stripe.setCoordinates(2*k, p);

		  if(k<pointcnt-1) // leave the last vertices
		  {
		    Vector3f v1 = new Vector3f(1.0f, 0.0f, 0.0f);
		    Vector3f v2 = new Vector3f(0.0f,
			                           (float)(((Range)scaling_v.elementAt(d)).normalize(
		                               ((Double)((Vector)points_v.elementAt(k+1)).elementAt(d)).doubleValue())-y),
									   (float) 1 / (float) (pointcnt-1) );
			v2.normalize();
			normal = new Vector3f();
			normal.cross(v1,v2);
			normal.normalize();

			stripe.setNormal(2*k, normal);
			stripe.setNormal(2*k+1, normal);
		  }
		  else // the last 2 verices of one strip get the same normals as their predecessor
		  {
			stripe.setNormal(2*k, normal);
			stripe.setNormal(2*k+1, normal);
		  }

		  // insert grid if gridon is set
		  if(style.showlines==true && d>0)
		  {
		    Point3d p1 = new Point3d( (double) (d-1)*(stripwidth+spacewidth)+stripwidth,
		  	                          ((Range)scaling_v.elementAt(d-1)).normalize(
		                              ((Double)((Vector)points_v.elementAt(k)).elementAt(d-1)).doubleValue()),
									  z);
		    Point3d p2 = new Point3d( x1, y, z);

            lines.setCoordinate((d-1)*pointcnt*2+k*2, p1);
            lines.setCoordinate((d-1)*pointcnt*2+k*2+1, p2);
		  }

		} // for

		stripe_v.addElement(stripe);

	  } // for

      for (int i=0; i<dimcnt; i++)
      {
        Shape3D shape = new Shape3D();
        shape.setGeometry( (TriangleStripArray) stripe_v.elementAt(i) );
        stripeShape_v.addElement( shape );
      }
	  if(style.showlines==true && dimcnt>0) lineShape.setGeometry(lines);

	}

   /**
    * Inserts the text2d-elements specifying the coordinate's range
    * into the BranchGroup
    */
    private void createAnnotation(int dimcnt, Vector scaling_v)
	{
	  TransformGroup tg;
	  Transform3D t;
      Text2D txt;
	  double stripwidth=0.0;
	  double spacewidth=0.0;

	  if(dimcnt>1)
	  {
	    stripwidth = (1-style.space_frac) / (double) dimcnt;
	    spacewidth = style.space_frac / ((double) dimcnt - 1.0);
	  }
	  else
	  {
	    stripwidth = 1.0;
	    spacewidth = 0.0;
	  }

	  for(int d=0; d<dimcnt; d++)
	  {
	    tg = new TransformGroup();
	    t = new Transform3D();

		t.setTranslation( new Vector3d((double)d*(stripwidth+spacewidth), -0.1, 1.01) );
		tg.setTransform(t);
		txt = new Text2D(""+((Range)scaling_v.elementAt(d)).getMin2(),
		                 new Color3f(0.9f, 1.0f, 1.0f), "System", 15, 1);

        tg.addChild(txt);
		addChild(tg);

	    tg = new TransformGroup();
	    t = new Transform3D();

		t.setTranslation( new Vector3d((double)d*(stripwidth+spacewidth), 1.0, 1.01) );
		tg.setTransform(t);
		txt = new Text2D(""+((Range)scaling_v.elementAt(d)).getMax2(),
		                 new Color3f(0.9f, 1.0f, 1.0f), "System", 15, 1);
        tg.addChild(txt);
		addChild(tg);

	  }

	  createTime();

	}

   /**
    * Inserts text3D-elements into the scene specifying the letters
    * for each variable.
    */
    private void createVariables(int dimcnt, Vector variablenames_v)
	{
	  TransformGroup tg, atg;
	  Transform3D t, at;
      Text2D txt;
	  Text3D txt3D;
	  double stripwidth=0.0;
	  double spacewidth=0.0;

	  if(dimcnt>1)
	  {
	    stripwidth = (1-style.space_frac) / (double) dimcnt;
	    spacewidth = style.space_frac / ((double) dimcnt - 1.0);
	  }
	  else
	  {
	    stripwidth = 1.0;
	    spacewidth = 0.0;
	  }

      Appearance app = new Appearance();
	  TransparencyAttributes ta = new TransparencyAttributes();
      ta.setTransparencyMode(ta.BLENDED);
      ta.setTransparency(0.4f);
      app.setTransparencyAttributes(ta);
      Color3f objColor = new Color3f(0.7f, 0.8f, 1.0f);
      Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
      app.setMaterial(new Material(objColor, black, objColor, black, 1.0f));

	  for(int d=0; d<dimcnt; d++)
	  {
        atg = new TransformGroup();
	    tg = new TransformGroup();
	    t = new Transform3D();
		at = new Transform3D();

		t.setTranslation( new Vector3d((double)d*(stripwidth+spacewidth), 0.0, 1.0) );
		tg.setTransform(t);

        Font f = new Font ("System", 1, 1);
		Font3D f3D = new Font3D(f, new FontExtrusion());

        txt3D = new Text3D(f3D, ""+((String)variablenames_v.elementAt(d)).toString() );

		Shape3D s3D = new Shape3D();
		s3D.setGeometry(txt3D);
        s3D.setAppearance(app);

        atg.addChild(s3D);
		at.setScale(0.1);
		atg.setTransform(at);

        tg.addChild(atg);

   		addChild(tg);
	  }

	}

   /**
    * Inserts the range of the time of the current dynamical system as
    * text3d.
    */
    private void createTime()
	{
	  TransformGroup tg, atg;
	  Transform3D t, at;
      Text2D txt;
	  Text3D txt3D;

      Appearance app = new Appearance();
	  TransparencyAttributes ta = new TransparencyAttributes();
      ta.setTransparencyMode(ta.BLENDED);
      ta.setTransparency(0.4f);
      app.setTransparencyAttributes(ta);
      Color3f objColor = new Color3f(0.3f, 1.0f, 0.9f);
      Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
      app.setMaterial(new Material(objColor, black, objColor, black, 1.0f));

      atg = new TransformGroup();
	  tg = new TransformGroup();
	  t = new Transform3D();
	  at = new Transform3D();

      t.rotY (1.57079);
      t.setScale(0.5);
	  t.setTranslation( new Vector3d(1.0, -0.05, 0.05) );
	  tg.setTransform(t);

      Font f = new Font ("System", 1, 1);
	  Font3D f3D = new Font3D(f, new FontExtrusion());

      txt3D = new Text3D(f3D, ""+parent.time.getMin2()+"  <- time");

	  Shape3D s3D = new Shape3D();
	  s3D.setGeometry(txt3D);
      s3D.setAppearance(app);

      atg.addChild(s3D);
	  at.setScale(0.1);
	  atg.setTransform(at);

      tg.addChild(atg);
	  addChild(tg);

      atg = new TransformGroup();
	  tg = new TransformGroup();
	  t = new Transform3D();
	  at = new Transform3D();

      t.rotY (1.57079);
      t.setScale(0.5);
	  t.setTranslation( new Vector3d(1.0, -0.05, 0.55) );
	  tg.setTransform(t);

      f = new Font ("System", 1, 1);
	  f3D = new Font3D(f, new FontExtrusion());

      txt3D = new Text3D(f3D, ""+((parent.time.getMax2()-parent.time.getMin2())/2.0) );

	  s3D = new Shape3D();
	  s3D.setGeometry(txt3D);
      s3D.setAppearance(app);

      atg.addChild(s3D);
	  at.setScale(0.1);
	  atg.setTransform(at);

      tg.addChild(atg);
	  addChild(tg);


      atg = new TransformGroup();
	  tg = new TransformGroup();
	  t = new Transform3D();
	  at = new Transform3D();

      t.rotY (1.57079);
      t.setScale(0.5);
	  t.setTranslation( new Vector3d(1.0, -0.05, 1.05) );
	  tg.setTransform(t);

      f = new Font ("System", 1, 1);
	  f3D = new Font3D(f, new FontExtrusion());

      txt3D = new Text3D(f3D, ""+parent.time.getMax2() );

	  s3D = new Shape3D();
	  s3D.setGeometry(txt3D);
      s3D.setAppearance(app);

      atg.addChild(s3D);
	  at.setScale(0.1);
	  atg.setTransform(at);

      tg.addChild(atg);
	  addChild(tg);

	}
   /**
    * Inserts the spheres, which are dragable with the mouse to change
    * the starting value of each variable
    */
    private void createController(Vector points_v, Vector scaling_v)
    {
	  double stripwidth=0.0;
	  double spacewidth=0.0;

  	  if(dimcnt>1)
	  {
	    stripwidth = (1-style.space_frac) / (double) dimcnt;
	    spacewidth = style.space_frac / ((double) dimcnt - 1.0);
	  }
	  else
	  {
	    stripwidth = 1.0;
	    spacewidth = 0.0;
	  }

      Appearance app = new Appearance();
      Color3f objColor = new Color3f(1.0f, 0.2f, 0.0f);

      Material mat = new Material(objColor, new Color3f(0.0f,0.0f,0.0f), objColor,
	                                              new Color3f(1.0f,1.0f,1.0f), 80.0f);
      app.setMaterial(mat);

      for (int i=0; i<dimcnt;i++)
      {
        BranchGroup bg = new BranchGroup();
        addChild(bg);

        TransformGroup tg = new TransformGroup();
         Transform3D t3d = new Transform3D();

		 double y = ((Range)scaling_v.elementAt(i)).normalize(
		               ((Double)((Vector)points_v.elementAt(0)).elementAt(i)).doubleValue());


         t3d.setTranslation(new Vector3d((double)i*(stripwidth+spacewidth)+stripwidth/2.0,
                            y,
                            0.0));

         tg.setTransform(t3d);
         tg.setCapability(Node.ENABLE_PICK_REPORTING);
         tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
         tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
        bg.addChild(tg);

        tg.addChild((new Sphere((float)stripwidth/2.0f, app)));

        PickTranslateBehavior2 mt = new PickTranslateBehavior2(bg,
                                                parent.canvas,
                                                new BoundingSphere(new Point3d(0.0, 1.0, 1.0), stripwidth/2.0),
                                                PickObject.USE_BOUNDS, (double)i*(stripwidth+spacewidth)+stripwidth/2.0,
                                                i,
                                                parent);
        bg.addChild(mt);

      }
    }

   /**
    * By calling this constructor the graph is created using the specified
    * data.
    */
    public EPCGraph(GraphOutput parent, int dimcnt, int pointcnt,
	                 Vector points_v, Style style, Vector scaling_v, Vector variablenames_v)
	{
	  this.parent = parent;
	  this.style = style;
	  this.dimcnt=dimcnt;
	  this.pointcnt=pointcnt;
	  createGeometries(dimcnt, pointcnt, points_v, scaling_v);
      createVariables(dimcnt, variablenames_v);
      createController(points_v, scaling_v);

      setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
      setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);

      if(style.showscaling==true) createAnnotation(dimcnt, scaling_v);

	  Appearance app = new Appearance();
	  Color3f color = new Color3f(1.0f, 1.0f, 1.0f);
	  ColoringAttributes ca = new ColoringAttributes();
	  ca.setColor(color);
	  app.setColoringAttributes(ca);

	  PolygonAttributes pa = new PolygonAttributes();
	  pa.setPolygonMode(pa.POLYGON_LINE);
	  pa.setCullFace(pa.CULL_NONE);
	  app.setPolygonAttributes(pa);

      for (int i=0; i<dimcnt; i++)
        ((Shape3D)stripeShape_v.elementAt(i)).setAppearance(app);

      stripe_bg = new BranchGroup();
      stripe_bg.setCapability(BranchGroup.ALLOW_DETACH);
      addChild(stripe_bg);
      for (int i=0; i<dimcnt; i++)
	    stripe_bg.addChild((Shape3D)stripeShape_v.elementAt(i));

	  if (style.showlines==true && dimcnt>1)
	  {
	    Appearance app1 = new Appearance();
	    Color3f color1 = new Color3f(1.0f, 1.0f, 1.0f);
	    ColoringAttributes ca1 = new ColoringAttributes();
	    ca1.setColor(color1);
	    app1.setColoringAttributes(ca1);

	    PolygonAttributes pa1 = new PolygonAttributes();
	    pa1.setPolygonMode(pa1.POLYGON_LINE);
	    pa1.setCullFace(pa1.CULL_NONE);
	    app1.setPolygonAttributes(pa1);
	    lineShape.setAppearance(app1);

 	    stripe_bg.addChild(lineShape);
	  }
	  setCapability(BranchGroup.ALLOW_DETACH);
      compile();

	}

   /**
    * By calling this constructor the graph is created using the specified
    * data.
    */
    public EPCGraph(GraphOutput parent, int dimcnt, int pointcnt,
	                Vector points_v, Vector scaling_v, Vector variablenames_v,
					Style style)
	{
	  this.parent = parent;
	  this.style = style;
	  this.dimcnt=dimcnt;
	  this.pointcnt=pointcnt;
	  createGeometries(dimcnt, pointcnt, points_v, scaling_v);
      createVariables(dimcnt, variablenames_v);
      createController(points_v, scaling_v);

      setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
      setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);

      for (int i=0; i<dimcnt; i++)
       ((Shape3D)stripeShape_v.elementAt(i)).setAppearance(style.getAppearance(i));

      stripe_bg = new BranchGroup();
      stripe_bg.setCapability(BranchGroup.ALLOW_DETACH);
      addChild(stripe_bg);
      for (int i=0; i<dimcnt; i++)
	    stripe_bg.addChild((Shape3D)stripeShape_v.elementAt(i));


      if(style.showscaling==true) createAnnotation(dimcnt, scaling_v);

	  if (style.showlines==true && dimcnt>1)
	  {
	    Appearance app1 = new Appearance();
	    Color3f color = new Color3f(1.0f, 1.0f, 1.0f);
	    ColoringAttributes ca = new ColoringAttributes();
	    ca.setColor(color);
	    app1.setColoringAttributes(ca);

	    PolygonAttributes pa = new PolygonAttributes();
	    pa.setPolygonMode(pa.POLYGON_LINE);
	    pa.setCullFace(pa.CULL_NONE);
	    app1.setPolygonAttributes(pa);

        LineAttributes lat = new LineAttributes();
        lat.setLineAntialiasingEnable(true);
	    lineShape.setAppearance(app1);

 	    stripe_bg.addChild(lineShape);
	  }
	  setCapability(BranchGroup.ALLOW_DETACH);
	  compile();
	}

} // end of class EPCGraph
