package graphoutput;

import com.sun.j3d.utils.behaviors.mouse.MouseRotate;
import com.sun.j3d.utils.behaviors.mouse.MouseZoom;
import com.sun.j3d.utils.behaviors.mouse.MouseTranslate;
import com.sun.j3d.utils.behaviors.picking.*;
import com.sun.j3d.utils.geometry.*;
import java.applet.*;
import java.awt.*;
import java.io.*;
import java.net.*;
import java.util.*;
import java.awt.BorderLayout;
import java.awt.event.*;
import javax.media.j3d.*;
import javax.vecmath.*;
import com.sun.j3d.utils.universe.*;
import com.sun.j3d.utils.image.*;
import DynSysInfo;
import NDimViewer;

/**
 * This class contains all output functionality and thus it appears as
 * an interface-class for all other classes
 */
public class GraphOutput extends Frame
implements ActionListener, ItemListener, WindowListener
{
   public static final int EPC = 1;
   public static final int LWW = 2;
   public static final int TDC = 3;
   public static final String TITLE_EPC = new String("Extruded Parallel Coordinates");
   public static final String TITLE_LWW = new String("Linking With Wings");
   public static final String TITLE_TDC = new String("Three-Dimensional Parallel Coordinates");

   Canvas3D canvas;
   Label    statusLine;
   Button   propButton;
   Button   initviewButton;
   Checkbox annotationCheckbox;
   Choice   techniqueChoice;
   Button   optionsdlgButton;
   Button   recalcButton;

   TransformGroup tg;
   TransformGroup btg;
   BranchGroup    graph;
   Transform3D initialCamera;

   public   int dimcnt;
   private  int pointcnt;
   public   int technique;
   public   Vector variablenames_v;
   public   Vector mapping_v;
   public   Vector scaling_v;
   public   Vector startvalue_v;
   public   Style style;
   private  Vector dataset_v;
   public   Vector3d coordBoxScale = new Vector3d(1.0f, 0.6f, 1.2f);
   public   Range time;
   private  DynSysInfo dsi = new DynSysInfo();
   private  EPCGraph shape_epc;
   private  LWWGraph shape_lww;
   private  TDCGraph shape_tdc;
   public   NDimViewer applet;
   private  BranchGroup cb_bg;

   public   OptionsDlg optionsdlg;

   /**
    * Creates an empty output-window containing an empty scene within a
    * Canvas3D and the status-line. The size of the window and the title
    * are specified by parameters
    */
   public GraphOutput(NDimViewer applet, String title, int width, int height)
   {
	 this.applet=applet;
     setTitle(title);
	 style = new Style();

     technique=EPC;
	 createUniverse();
     setSize(width, height);
     updateAnnotation();
   }

   /**
    * Creates an empty output-window containing an empty scene within a
    * Canvas3D and the status-line
    */
   public GraphOutput(NDimViewer applet)
   {
     this.applet=applet;
     setTitle("Output Window");
     style = new Style();

	 technique=EPC;
	 createUniverse();
   	 setSize(640, 480);
	 pack();
	 updateAnnotation();
   }

   /**
    * Reacts on the actions performed by the user
	*/
   public void actionPerformed(ActionEvent e)
   {
	 if(e.getActionCommand().equals("Reset Camera"))
	 {
	   initCameraPosition();
	 }
	 else if(e.getActionCommand().equals("Properties"))
	 {
	   DSInfoDlg dlg = new DSInfoDlg(dsi);
	   dlg.setVisible(true);
	 }
	 else if(e.getActionCommand().equals("Options"))
	 {
	   if(optionsdlg == null)
	   {
	     optionsdlg = new OptionsDlg(this);
	     optionsdlg.setVisible(true);
	   }
	 }
	 else if(e.getActionCommand().equals("RECALC"))
	 {
	   //calc new points
       Vector save_scaling_v = (Vector) scaling_v.clone();

       if(technique == EPC) shape_epc.removeGeometry();
       else if(technique == LWW) shape_lww.removeGeometry();
       else if(technique == TDC) shape_tdc.removeGeometry();

       applet.iterate(startvalue_v);
       scaling_v = (Vector) save_scaling_v.clone();

       if(technique == EPC)
         shape_epc.updateGeometry(dataset_v, scaling_v);
       else if(technique == LWW)
         shape_lww.updateGeometry(dataset_v, scaling_v);
       else if(technique == TDC)
         shape_tdc.updateGeometry(dataset_v, scaling_v);

     }
   }

   /**
    * Is called if the state of the Checkbox is changed
	*/
   public void itemStateChanged(ItemEvent e)
   {
     if(((String)e.getItem())=="Annotation")
	 {
       if(e.getStateChange()==ItemEvent.SELECTED)
	   {
	     setAnnotation(true);
	   }
	   else
	   {
	     setAnnotation(false);
	   }
	 }
     else if(((String)e.getItem())==TITLE_EPC && technique!=EPC)
     {
	   style.marker_v = new Vector();
	   setTechnique(EPC);
     }
	 else if(((String)e.getItem())==TITLE_LWW && technique!=LWW)
     {
	   style.marker_v = new Vector();
	   setTechnique(LWW);
     }
	 else if(((String)e.getItem())==TITLE_TDC && technique!=TDC)
     {
	   style.marker_v = new Vector();
	   setTechnique(TDC);
     }
   }

   /**
    *  Resets the view to be able to get receive a new
	*  dataset_v
	*/
   public void initOutput(int dimcnt, Vector variablenames_v, Vector mapping_v, Range time)
   {
     this.dimcnt=dimcnt;
	 this.pointcnt=0;
	 this.variablenames_v = (Vector)variablenames_v.clone();
	 this.mapping_v = (Vector)mapping_v.clone();
	 this.time = time;
	 dataset_v = new Vector(); // empty the data-vector
	 scaling_v = new Vector(); // empty the scaling-vector
	 for (int i=0; i<dimcnt; i++) // init the scaling-vector
	   scaling_v.addElement(new Range());
	 // + clear view
   }

   /**
    *  Adds 1 point to the list of points.
	*  Return false, if the dimension is not equal to the
    *  previously set
	*/
   public boolean addPoint(Vector newpoint)
   {
     if(dataset_v.size()==0) startvalue_v = (Vector) newpoint.clone();

     if(newpoint.size()==dimcnt)
	 {
	  Vector p = (Vector) newpoint.clone();
	  dataset_v.addElement( p );
	  for(int i=0; i<dimcnt; i++)
          ((Range)scaling_v.elementAt(i)).checkRange( ((Double)p.elementAt(i)).doubleValue() );
	  pointcnt++;
	  return true;
	 }
	 return false;
   }

   /**
    *  Tells GraphOutput to start the rendering
	*/
   public void display()
   {
    if(graph.numChildren()!=0)
	{
	  for (int i=graph.numChildren()-1; i>=0; i--)
       ((BranchGroup)graph.getChild(i)).detach();
    }
    applyMapping(mapping_v);


    // add the primitives
	setStatusLine("Calculating Graph ...");
	switch(technique)
	{
	  case EPC:
        shape_epc = new EPCGraph(this, dimcnt, dataset_v.size(), dataset_v, scaling_v, variablenames_v, style);
        graph.addChild(shape_epc);
        break;

      case LWW:
        shape_lww = new LWWGraph(this, dimcnt, dataset_v.size(), dataset_v, scaling_v, variablenames_v, style);
        graph.addChild(shape_lww);
        break;

      case TDC:
        shape_tdc = new TDCGraph(this, dimcnt, dataset_v.size(), dataset_v, scaling_v, variablenames_v, style);
        graph.addChild(shape_tdc);
        break;

	  default:
	    break;
	 }

	 style.redrawCuttingPlanes(graph);
	 style.redrawMarkers(graph);

	 show();
     setStatusLine("ready");
   }

   /**
    *  Maps the current dataset to the specified order of coordinates
	*  and sets variablenames_v equal to mapping_v
	*/
   private void applyMapping(Vector mapping_v)
   {
     boolean ok = true;
     Vector newdataset_v = new Vector();
	 Vector order_v = new Vector();
	 Vector newscaling_v = new Vector();
	 // check if mapping is necessary and possible
	 if(mapping_v.equals(variablenames_v) && mapping_v.size()!=variablenames_v.size()) return;

	 for (int i=0; i<mapping_v.size(); i++)
	 {
	   int pos = variablenames_v.indexOf( (String)mapping_v.elementAt(i) );
	   // cancel mapping if mapping_v contains a wrong variablename
	   if(pos==-1) ok = false;
	   order_v.addElement(new Integer(pos));

	   newscaling_v.addElement( (Range)scaling_v.elementAt(pos) );

	 }

	 if(ok==false) return;

	 for (int i=0; i<dataset_v.size(); i++)
	 {
       Vector point = new Vector();
       for (int j=0; j<mapping_v.size(); j++)
	   {
	     point.addElement( (Double)((Vector)dataset_v.elementAt(i)).elementAt(((Integer)order_v.elementAt(j)).intValue()) );
	   }
	   newdataset_v.addElement(point);
	 }

	 // make the new ordered dataset to the current
	 dataset_v = (Vector) newdataset_v.clone();
	 variablenames_v = (Vector) mapping_v.clone();
     scaling_v = (Vector) newscaling_v.clone();
   }

   /**
    *  Shows or hides the annotation of the current style
    */
   public void setAnnotation(boolean visible)
   {
     style.showscaling = visible;
	 display();
   }

   /**
    *  Updates the Annotation-Checkbox according to the
	*  current state of showscaling in style
    */
   public void updateAnnotation()
   {
     annotationCheckbox.setState(style.showscaling);
   }

   /**
    *  Set a specific arrangement of the axes
	*/
   public void setMapping(Vector mapping_v)
   {
     this.mapping_v = (Vector)mapping_v.clone();
	 display();
   }

   /**
    *  Get the current arrangement of the axes
	*/
   public Vector getMapping()
   {
     return (Vector)mapping_v.clone();
   }

   /**
    *  Set the scaling for each axis
	*/
   public void setScaling(Vector scaling_v)
   {
     this.scaling_v=(Vector)scaling_v.clone();
   }

   /**
    *  Sets the size of the coordinate box
	*  default value is (0.5f, 0.3f, 0.6f)
	*/
   public void setBoxSize(double x, double y, double z)
   {
     coordBoxScale = new Vector3d ( x, y, z );
	 Transform3D t = new Transform3D();
	 t.setTranslation( new Vector3d(-coordBoxScale.x/2.0,
		                            -coordBoxScale.y/2.0,
									-coordBoxScale.z/2.0) );
	 t.setScale ( coordBoxScale );
	 btg.setTransform(t);
   }

   /**
    *  Sets the technique for nDim-Displaying
	*/
   public void setTechnique(int technique)
   {
     this.technique = technique;
	 if(graph.numChildren()!=0)
	 {
	   display();
	 }
   }

   /**
    *  Sets the current style
	*/
   public void setStyle(Style style)
   {
     this.style = style;
	 updateAnnotation();
   }

   /**
    *  Returns the current style
   	*/
   public Style getStyle()
   {
     return style;
   }


   /**
    *  Returns the camera to its original position
	*/
   private void initCameraPosition()
   {
     tg.setTransform( initialCamera );
   }

   /**
    * Writes information about the epc-marker into the
    * statusline
    */
   public void setHighlightInfo(int nr, double pos)
   {
     String s = new String("");

     int set_nr = (int)(pos*(double)dataset_v.size());
     if (set_nr < dataset_v.size())
     {
       s = new String(" State: [");
       for(int i=0; i<((Vector)dataset_v.elementAt(set_nr)).size(); i++)
       {
         if(i>0) s = s + ", ";

         Integer n = new Integer( (new Double((1000.0*((Double)((Vector)dataset_v.elementAt(set_nr)).elementAt(i)).doubleValue()))).intValue() );
         s = s + (n.doubleValue() / 1000.0);
       }
       s = s + "]";
     }

     setStatusLine("Nr: "+nr+", at time "+time.fitIntoRange(pos)+s);
   }


   /**
    *  Sets the text in the statusline of GraphOutput
    */
   public void setStatusLine(String text)
   {
     statusLine.setText( text );
   }

   /**
    * Delivers an instance of DynSysInfo with the information
    * about the viewed Dyn.Sys.
    */
   public void setDynSysInfo(DynSysInfo dsi)
   {
     this.dsi=dsi;
   }

   /**
    * Updates the Vector startvalue_v and displays the change in
    * the statusline
    */
   public void setSelStartValue(int dim, double value)
   {
     String s = new String("New Startpoint: (");

     double rv = ((Range)scaling_v.elementAt(dim)).fitIntoRange(value);

     startvalue_v.set(dim, new Double(rv));

     double w = (new Integer( (new Double((1000.0*((Double)startvalue_v.elementAt(0)).doubleValue()))).intValue() )).doubleValue()/1000.0;
     s = s + w;
     for(int i=1; i<startvalue_v.size(); i++)
     {
       w = (new Integer( (new Double((1000.0*((Double)startvalue_v.elementAt(i)).doubleValue()))).intValue() )).doubleValue()/1000.0;
       s = s + ", " + w;
     }

     setStatusLine(s + ")");
   }

   /**
    * Updates the Vector startvalue_v and displays the change in
    * the statusline
    */
   public void setSelStartValueTDC(int plane_no, double a, double b)
   {
     String s = new String("New Startpoint: (");

     double ra = ((Range)scaling_v.elementAt(plane_no*2)).fitIntoRange(a);
     double rb = 0.0;
     if(plane_no*2+1 < dimcnt)
       rb = ((Range)scaling_v.elementAt(plane_no*2+1)).fitIntoRange(b);
     else rb=ra;

     startvalue_v.set(plane_no*2, new Double(ra));
     if(plane_no*2+1<dimcnt)
       startvalue_v.set(plane_no*2+1, new Double(rb));

     double w = (new Integer( (new Double((1000.0*((Double)startvalue_v.elementAt(0)).doubleValue()))).intValue() )).doubleValue()/1000.0;
     s = s + w;
     for(int i=1; i<startvalue_v.size(); i++)
     {
       w = (new Integer( (new Double((1000.0*((Double)startvalue_v.elementAt(i)).doubleValue()))).intValue() )).doubleValue()/1000.0;
       s = s + ", " + w;
     }

     setStatusLine(s + ")");
   }


   /**
    * Updates the Vector startvalue_v and displays the change in
    * the statusline
    */
   public void setSelStartValue(double x, double y)
   {

     double rx = ((Range)scaling_v.elementAt(0)).fitIntoRange(x);
     double ry = ((Range)scaling_v.elementAt(1)).fitIntoRange(y);

     startvalue_v.set(0, new Double(rx));
     startvalue_v.set(1, new Double(ry));

     showStartValue();
   }

   /**
    * Displays the current startvalue_v in the statusline
    */
   public void showStartValue()
   {
     String s = new String("New Startpoint: (");

     double w = (new Integer( (new Double((1000.0*((Double)startvalue_v.elementAt(0)).doubleValue()))).intValue() )).doubleValue()/1000.0;
     s = s + w;
     for(int i=1; i<startvalue_v.size(); i++)
     {
       w = (new Integer( (new Double((1000.0*((Double)startvalue_v.elementAt(i)).doubleValue()))).intValue() )).doubleValue()/1000.0;
       s = s + ", " + w;
     }

     setStatusLine(s + ")");
   }

   /**
    * Removes the CoordinateBox from the Scene-graph
    */
   public void hideCoordinateBox()
   {
     if (cb_bg != null)
      cb_bg.detach();
   }

   /**
    * Puts the CoordinateBox into the Scene-graph
    */
   public void showCoordinateBox()
   {
     try
     {
       btg.addChild(cb_bg);
     } catch (javax.media.j3d.MultipleParentException e) { }

   }

   /**
    * Stores the position and orientation of the LWW-trajectory
    * at the given pos [0;1] into p and v
    */
   public void getLWWOrientation(double pos, Point3d p, Vector3d v)
   {
     Point3d p1 = new Point3d();
     int ds_pos = (int)(pos*(double)(dataset_v.size()-1));

     p.x = ((Range)scaling_v.elementAt(0)).normalize(
           ((Double)((Vector)dataset_v.elementAt(ds_pos)).elementAt(0)).doubleValue()
           );
     p.y = 0.5;
     p.z = ((Range)scaling_v.elementAt(1)).normalize(
           ((Double)((Vector)dataset_v.elementAt(ds_pos)).elementAt(1)).doubleValue()
           );

     if(ds_pos == dataset_v.size()-1) // for calculating the normalvector the preceding
     {                                // point has to be taken
       p1.x = ((Range)scaling_v.elementAt(0)).normalize(
              ((Double)((Vector)dataset_v.elementAt(ds_pos-1)).elementAt(0)).doubleValue()
              );
       p1.y = 0.5;
       p1.z = ((Range)scaling_v.elementAt(1)).normalize(
              ((Double)((Vector)dataset_v.elementAt(ds_pos-1)).elementAt(1)).doubleValue()
              );

       v.x = p.x - p1.x;
       v.y = 0.0;
       v.z = p.z - p1.z;
     }
     else  // take the next point to calculate the normal vector
     {
       p1.x = ((Range)scaling_v.elementAt(0)).normalize(
              ((Double)((Vector)dataset_v.elementAt(ds_pos+1)).elementAt(0)).doubleValue()
              );
       p1.y = 0.5;
       p1.z = ((Range)scaling_v.elementAt(1)).normalize(
              ((Double)((Vector)dataset_v.elementAt(ds_pos+1)).elementAt(1)).doubleValue()
              );

       v.x = p1.x - p.x;
       v.y = 0.0;
       v.z = p1.z - p.z;
     }
     v.normalize();
   }

   /**
    * Returns the current point at the specified position
    */
   public Vector getTDCPosition(double pos)
   {
     Vector result_v = new Vector();

	 int ds_pos = (int)(pos*(double)(dataset_v.size()-1));

     result_v = (Vector)((Vector) (dataset_v.elementAt(ds_pos))).clone();

     return result_v;
   }



   /**
    *  Create the universe
    */
   private void createUniverse()
   {

     // initialize graph
	 graph = new BranchGroup();

     // Set the layout manager
     setLayout(new BorderLayout());

     // first add the interaction components
	 Panel p = new Panel(new BorderLayout());
	 statusLine = new Label("ready");
	 statusLine.setBackground(Color.white);
	 p.add("Center", statusLine);

	 Panel bp = new Panel(new FlowLayout());

	 techniqueChoice = new Choice();
	 techniqueChoice.add(TITLE_EPC);
	 techniqueChoice.add(TITLE_LWW);
	 techniqueChoice.add(TITLE_TDC);
	 techniqueChoice.addItemListener(this);
	 bp.add(techniqueChoice);
	 annotationCheckbox = new Checkbox("Annotation", false);
	 annotationCheckbox.addItemListener(this);
	 bp.add(annotationCheckbox);
	 Button initviewButton = new Button("Reset Camera");
     initviewButton.addActionListener(this);
 	 bp.add(initviewButton);
	 propButton = new Button("Properties");
     propButton.addActionListener(this);
	 bp.add(propButton);

     optionsdlgButton = new Button("Options");
	 optionsdlgButton.addActionListener(this);
	 bp.add(optionsdlgButton);

     recalcButton = new Button("RECALC");
	 recalcButton.addActionListener(this);
	 bp.add(recalcButton);

	 p.add("East", bp);
	 add("South",p);

     // Create the 3D canvas
     canvas = new Canvas3D(null);
     add("Center", canvas);

     addWindowListener(this);

     // Create the scene branch graph
     BranchGroup scene = createSceneGraph();

     // Create the Universe, the Locale, and the view branch graph
     SimpleUniverse u = new SimpleUniverse(canvas);

     // This will move the ViewPlatform back a bit so the
     // objects in the scene can be viewed.
     u.getViewingPlatform().setNominalViewingTransform();

     u.addBranchGraph(scene);
   }

   /**
    *  Create the scenegraph for this program.
    */
   private BranchGroup createSceneGraph()
   {
        // Define colors
        Color3f white = new Color3f(1.0f, 1.0f, 1.0f);
        Color3f black = new Color3f(0.0f, 0.0f, 0.0f);
        Color3f red   = new Color3f(0.80f, 0.20f, 0.2f);
        Color3f ambientRed = new Color3f(0.2f, 0.05f, 0.0f);
        Color3f ambient = new Color3f(0.2f, 0.2f, 0.2f);
        Color3f diffuse = new Color3f(0.7f, 0.7f, 0.7f);
        Color3f specular = new Color3f(0.7f, 0.7f, 0.7f);
        Color3f bgColor = new Color3f(0.0f, 0.0f, 0.0f);

        // Create the branch group
        BranchGroup branchGroup = new BranchGroup();

        // Create a Transformgroup to scale all objects so they
        // appear in the scene.
        TransformGroup objScale = new TransformGroup();
		objScale.clearCapability( TransformGroup.ENABLE_PICK_REPORTING );
        Transform3D t3d = new Transform3D();
        t3d.setScale(1.0);
        objScale.setTransform(t3d);
        branchGroup.addChild(objScale);

        // Create the bounding leaf node
        BoundingSphere bounds =
           new BoundingSphere(new Point3d(0.0,0.0,0.0), 100.0);
        BoundingLeaf boundingLeaf = new BoundingLeaf(bounds);

        // Set up the background
        Background bg = new Background(bgColor);
        bg.setApplicationBounds(bounds);
        objScale.addChild(bg);

        // Create the ambient light
        AmbientLight ambLight = new AmbientLight(white);
        ambLight.setInfluencingBounds(bounds);
        objScale.addChild(ambLight);

        // Create the directional light
        Vector3f dir = new Vector3f(-1.0f, -1.0f, -1.0f);
        DirectionalLight dirLight = new DirectionalLight(white, dir);
        dirLight.setInfluencingBounds(bounds);
        objScale.addChild(dirLight);

        // Create the point light
        Point3f pos = new Point3f(0.0f, 0.0f, 0.0f);
		Point3f att = new Point3f(1.0f, 1.0f, 1.0f);
        PointLight ptLight = new PointLight(true, white, pos, att);
        ptLight.setInfluencingBounds(bounds);
        objScale.addChild(ptLight);

        // Create the transformation node tg
		// which is used for rotation
        tg = new TransformGroup();
        tg.setCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        tg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		tg.setCapability(TransformGroup.ENABLE_PICK_REPORTING );
		Transform3D x1 = new Transform3D();
		initialCamera = new Transform3D(x1);
		tg.setTransform(x1);

        // insert into scene
        objScale.addChild(tg);

        // create basic transformation node btg
		btg = new TransformGroup();
        btg.clearCapability(TransformGroup.ALLOW_TRANSFORM_READ);
        btg.setCapability(TransformGroup.ALLOW_TRANSFORM_WRITE);
		btg.clearCapability(TransformGroup.ENABLE_PICK_REPORTING);
		btg.setCapability(TransformGroup.ALLOW_CHILDREN_WRITE);
		btg.setCapability(TransformGroup.ALLOW_CHILDREN_EXTEND);
		Transform3D x2 = new Transform3D();
		x2.setTranslation( new Vector3d(-coordBoxScale.x/2.0,
		                                -coordBoxScale.y/2.0,
										-coordBoxScale.z/2.0) );
		x2.setScale( coordBoxScale );
		btg.setTransform(x2);

        // insert into scene
        tg.addChild(btg);

		// insert objects into btg
		cb_bg = new BranchGroup();
		cb_bg.setCapability(BranchGroup.ALLOW_DETACH);
		cb_bg.addChild(new CoordinateBox().getChild());
        btg.addChild(cb_bg);
		graph = new BranchGroup();
		graph.setCapability(BranchGroup.ALLOW_CHILDREN_EXTEND);
		graph.setCapability(BranchGroup.ALLOW_CHILDREN_READ);
		graph.setCapability(BranchGroup.ALLOW_CHILDREN_WRITE);
		graph.setCapability(BranchGroup.ALLOW_DETACH);
		btg.addChild( graph );

        // Create the rotate behavior nodes
        PickRotateBehavior behavior1 = new PickRotateBehavior(branchGroup, canvas, bounds);
        objScale.addChild(behavior1);
	    // Create the zoom behavior nodes
        PickZoomBehavior behavior3 = new PickZoomBehavior(branchGroup, canvas, bounds);
        objScale.addChild(behavior3);

        // Let Java 3D perform optimizations on this scene graph.
        branchGroup.compile();

        return branchGroup;
    }


    public void windowActivated(WindowEvent e) {}
    public void windowClosed(WindowEvent e) {}
    public void windowClosing(WindowEvent e)
	{
	  if (optionsdlg!=null)
	  {
	    optionsdlg.setVisible(false);
	    optionsdlg.dispose();
	  }
      setVisible(false);
	  dispose();
	}
    public void windowDeactivated(WindowEvent e) {}
    public void windowDeiconified(WindowEvent e) {}
    public void windowIconified(WindowEvent e) {}
    public void windowOpened(WindowEvent e) {}


}
